@@ -39,6 +39,8 @@ PARTICLE_MODEL = "particle_model"
 regist_config(PARTICLE_MODEL, {})
 DECAY_MODEL = "decay_model"
 regist_config(DECAY_MODEL, {})
+DECAY_CHAIN_MODEL = "decay_chain_model"
+regist_config(DECAY_CHAIN_MODEL, {})
 
 
 def register_particle(name=None, f=None):
@@ -90,6 +92,31 @@ def register_decay(name=None, num_outs=2, f=None):
     return regist(f)
 
 
+def register_decay_chain(name=None, f=None):
+    """register a decay model
+
+    :params name: model name used in configuration
+    :params f: Model class
+    """
+
+    def regist(g):
+        if name is None:
+            my_name = g.__name__
+        else:
+            my_name = name
+        config = get_config(DECAY_CHAIN_MODEL)
+        id_ = my_name
+        if id_ in config:
+            warnings.warn("Override deccay model {}".format(my_name))
+        config[id_] = g
+        g.model_name = my_name
+        return g
+
+    if f is None:
+        return regist
+    return regist(f)
+
+
 regist_particle = register_particle
 regist_decay = register_decay
 
@@ -166,6 +193,19 @@ def get_decay(core, outs, **kwargs):
     return get_config(DECAY_MODEL)[id_](core, outs, **new_kwargs)
 
 
+def get_decay_chain(decays, **kwargs):
+    """method for getting decay of model"""
+    decay_params = {}
+    for i in decays:
+        decay_params.update(getattr(i, "decay_chain_params", {}))
+
+    new_kwargs = {**decay_params, **kwargs}
+
+    model = new_kwargs.pop("model", "default")
+
+    return get_config(DECAY_CHAIN_MODEL)[model](decays, **new_kwargs)
+
+
 def data_device(data):
     def get_device(dat):
         if hasattr(dat, "device"):
@@ -485,6 +525,7 @@ class HelicityDecay(AmpDecay):
         allow_cc=True,
         ls_list=None,
         barrier_factor_norm=False,
+        params_polar=None,
         **kwargs
     ):
         super(HelicityDecay, self).__init__(*args, **kwargs)
@@ -501,6 +542,7 @@ class HelicityDecay(AmpDecay):
         self.ls_list = None
         if ls_list is not None:
             self.ls_list = tuple([tuple(i) for i in ls_list])
+        self.params_polar = params_polar
 
     def check_valid_jp(self):
         if len(self.get_ls_list()) == 0:
@@ -529,7 +571,9 @@ class HelicityDecay(AmpDecay):
     def init_params(self):
         self.d = 3.0
         ls = self.get_ls_list()
-        self.g_ls = self.add_var("g_ls", is_complex=True, shape=(len(ls),))
+        self.g_ls = self.add_var(
+            "g_ls", is_complex=True, polar=self.params_polar, shape=(len(ls),)
+        )
         try:
             self.g_ls.set_fix_idx(fix_idx=0, fix_vals=(1.0, 0.0))
         except Exception as e:
@@ -859,15 +903,18 @@ class AngSam3Decay(AmpDecay, AmpBase):
         return ret
 
 
-class DecayChain(BaseDecayChain, AmpBase):
-    """A list of Decay as a chain decay"""
-
+class AmpDecayChain(BaseDecayChain, AmpBase):
     def __init__(self, *args, is_cp=False, **kwargs):
         self.is_cp = is_cp
-        super(DecayChain, self).__init__(*args, **kwargs)
+        super(AmpDecayChain, self).__init__(*args, **kwargs)
         self.aligned = True
         self.need_amp_particle = True
 
+
+@register_decay_chain("default")
+class DecayChain(AmpDecayChain):
+    """A list of Decay as a chain decay"""
+
     def init_params(self, name=""):
         self.total = self.add_var(
             name + "total", is_complex=True, is_cp=self.is_cp, shape=[1]
@@ -1106,6 +1153,7 @@ class DecayGroup(BaseDecayGroup, AmpBase):
             chains = [DecayChain(i) for i in chains]
         super(DecayGroup, self).__init__(chains)
         self.not_full = False
+        self.polarization = getattr(self.top, "polarization", "none")
         # self.init_params()
 
     def init_params(self, name=""):
@@ -1118,6 +1166,14 @@ class DecayGroup(BaseDecayGroup, AmpBase):
                 if j not in inited_set:
                     j.init_params()
                     inited_set.add(j)
+        if self.polarization == "vector":
+            print("add polarization vector")
+            if self.top.J == 0.5:
+                self.polarization_vector = [
+                    self.top.add_var("polarization_px"),
+                    self.top.add_var("polarization_py"),
+                    self.top.add_var("polarization_pz"),
+                ]
 
     def get_factor_variable(self):
         ret = []
@@ -1155,6 +1211,7 @@ class DecayGroup(BaseDecayGroup, AmpBase):
                     data_c, data_p, base_map=base_map, all_data=data
                 )
                 ret.append(amp)
+                # print(decay_chain, amp[:10])
         ret = tf.reduce_sum(ret, axis=0)
         return ret
 
@@ -1167,25 +1224,28 @@ class DecayGroup(BaseDecayGroup, AmpBase):
         chain_maps = self.get_chains_map(used_chains)
         base_map = self.get_base_map()
         ret = []
-        for chains in chain_maps:
-            for decay_chain in chains:
-                chain_topo = decay_chain.standard_topology()
-                found = False
-                for i in data_decay.keys():
-                    if i == chain_topo:
-                        data_decay_i = data_decay[i]
-                        found = True
-                        break
-                if not found:
-                    raise KeyError("not found {}".format(chain_topo))
-                data_c = rename_data_dict(data_decay_i, chains[decay_chain])
-                data_p = rename_data_dict(data_particle, chains[decay_chain])
-                # print("$$$$$",data_c)
-                # print("$$$$$",data_p)
-                amp = decay_chain.get_m_dep(
-                    data_c, data_p, base_map=base_map, all_data=data
-                )
-                ret.append(amp)
+        for decay_chain in used_chains:
+            for chains in chain_maps:
+                if str(decay_chain) in [str(i) for i in chains]:
+                    maps = chains[decay_chain]
+                    break
+            chain_topo = decay_chain.standard_topology()
+            found = False
+            for i in data_decay.keys():
+                if i == chain_topo:
+                    data_decay_i = data_decay[i]
+                    found = True
+                    break
+            if not found:
+                raise KeyError("not found {}".format(chain_topo))
+            data_c = rename_data_dict(data_decay_i, maps)
+            data_p = rename_data_dict(data_particle, maps)
+            # print("$$$$$",data_c)
+            # print("$$$$$",data_p)
+            amp = decay_chain.get_m_dep(
+                data_c, data_p, base_map=base_map, all_data=data
+            )
+            ret.append(amp)
         # ret = tf.reduce_sum(ret, axis=0)
         return ret
 
@@ -1197,38 +1257,104 @@ class DecayGroup(BaseDecayGroup, AmpBase):
         chain_maps = self.get_chains_map(used_chains)
         base_map = self.get_base_map()
         ret = []
-        for chains in chain_maps:
-            for decay_chain in chains:
-                chain_topo = decay_chain.standard_topology()
-                found = False
-                for i in data_decay.keys():
-                    if i == chain_topo:
-                        data_decay_i = data_decay[i]
-                        found = True
-                        break
-                if not found:
-                    raise KeyError("not found {}".format(chain_topo))
-                data_c = rename_data_dict(data_decay_i, chains[decay_chain])
-                data_p = rename_data_dict(data_particle, chains[decay_chain])
-                amp = decay_chain.get_angle_amp(
-                    data_c, data_p, base_map=base_map, all_data=data
-                )
-                ret.append(amp)
+        for decay_chain in used_chains:
+            for chains in chain_maps:
+                if str(decay_chain) in [str(i) for i in chains]:
+                    maps = chains[decay_chain]
+                    break
+            chain_topo = decay_chain.standard_topology()
+            found = False
+            for i in data_decay.keys():
+                if i == chain_topo:
+                    data_decay_i = data_decay[i]
+                    found = True
+                    break
+            if not found:
+                raise KeyError("not found {}".format(chain_topo))
+            data_c = rename_data_dict(data_decay_i, maps)
+            data_p = rename_data_dict(data_particle, maps)
+            amp = decay_chain.get_angle_amp(
+                data_c, data_p, base_map=base_map, all_data=data
+            )
+            ret.append(amp)
         # ret = tf.reduce_sum(ret, axis=0)
         return amp
 
+    @functools.lru_cache()
+    def get_swap_factor(self, key):
+        factor = 1.0
+        for i, j in zip(self.identical_particles, key[1]):
+            p = self.get_particle(i[0])
+            if int(p.J * 2) % 2 == 0:
+                continue
+            for m, n in zip(i, j):
+                if m != n:
+                    factor *= -1.0
+        return factor
+
+    def get_amp2(self, data):
+        amp = self.get_amp(data)
+        id_swap = data.get("id_swap", {})
+        for k, v in id_swap.items():
+            new_data = {**data, **v}
+            factor = self.get_swap_factor(k)
+            amp_swap = factor * self.get_amp(new_data)
+            # print(k, amp, amp_swap)
+            amp = amp + amp_swap
+        return amp
+
     def sum_amp(self, data, cached=True):
         """
         calculat the amplitude modular square
         """
         if not cached:
             data = simple_deepcopy(data)
-        amp = self.get_amp(data)
+        if self.polarization != "none":
+            return self.sum_amp_polarization(data)
+        amp = self.get_amp2(data)
         amp2s = tf.math.real(amp * tf.math.conj(amp))
         idx = list(range(1, len(amp2s.shape)))
         sum_A = tf.reduce_sum(amp2s, idx)
         return sum_A
 
+    def sum_amp_polarization(self, data):
+        """
+        sum amplitude suqare with density _get_cg_matrix
+
+        .. math::
+            P = \\sum_{m, m', \\cdots } A_{m, \\cdots}  \\rho_{m, m'} A^{*}_{m', \\cdots}
+
+        """
+
+        amp = self.get_amp(data)
+        amp = tf.reshape(
+            amp, (amp.shape[0], amp.shape[1], -1)
+        )  # (i, la, lb lc ld ...)
+        na, nl = amp.shape[1], amp.shape[2]
+        rho = self.get_density_matrix()
+        amp = tf.reshape(amp, (-1, na, 1, nl))
+        amp_c = tf.reshape(
+            tf.math.conj(amp), (-1, na, nl)
+        )  # (i, la, lb lc ld ...)
+        sum_A = (
+            tf.reduce_sum(amp * tf.reshape(rho, (na, na, 1)), axis=1) * amp_c
+        )
+        return tf.reduce_sum(tf.math.real(sum_A), axis=[1, 2])
+
+    def get_density_matrix(self):
+        if self.polarization == "vector":
+            px, py, pz = [i() for i in self.polarization_vector]
+            zeros = tf.zeros_like(px)
+            ones = tf.ones_like(px)
+            rho00 = tf.complex(ones + pz, zeros)
+            rho11 = tf.complex(ones - pz, zeros)
+            rho01 = tf.complex(px, -py)
+            rho10 = tf.complex(px, py)
+            ret = 0.5 * tf.stack([[rho00, rho01], [rho10, rho11]])
+            # print(ret)
+            return ret
+        raise NotImplementedError
+
     # @simple_cache_fun
     def amp_index(self, gen=None, base_map=None):
         if base_map is None:
