@@ -6,6 +6,7 @@ going into every modules. There are some example files where you can figure out
 """
 import os
 import time
+import warnings
 
 import matplotlib.pyplot as plt
 import numpy as np
@@ -191,7 +192,83 @@ def calPWratio(params, POLAR=True):
 '''
 
 
-def cal_hesse_error(fcn, params={}, check_posi_def=True, save_npy=True):
+def force_pos_def_minuit2(inv_he):
+    """
+    force positive defined of error matrix
+
+        from minuit2 https://github.com/root-project/root/blob/master/math/minuit2/sec/MnPosDef.cxx
+    """
+    inv_he = np.array(inv_he)
+    finfo = np.info(inv_he.dtype)
+    # print(finfo, inv_he.dtype)
+    if finfo is not None:
+        eps = finfo.eps
+    else:
+        eps = 2.220446049250313e-16  # float64
+    eps2 = 2 * np.sqrt(eps)
+
+    diag = np.diagonal(inv_he)
+    diag_i = np.ones_like(diag)
+    dgmin = np.min(diag)
+    if dgmin > 0:
+        return inv_he
+    epspdf = max(1e-6, eps2)
+    dg = 0.5 + epspdf - dgmin
+    inv_he += np.diag(diag_i * dg)
+    warnings.warn("Added to diagonal of Error Matrix a value {}".format(dgmin))
+    diag2 = np.diagonal(inv_he)
+    diag2 = np.where(diag2 < 0, 1, diag2)
+    e = 1 / np.sqrt(diag2)
+    p = inv_he.copy() / diag[:, None] / diag[None, :]
+    p = np.triu(p, 0)
+    e, v = np.linalg.eig(p)
+    pmin = np.min(e)
+    pmax = max(np.max(np.abs(e)), 1)
+    if pmin > epspdf * pmax:
+        return inv_he
+
+    padd = 0.001 * pmax - pmin
+    inv_he += np.diag(np.diag(inv_he) * padd)
+    warnings.warn(
+        "Matrix forced pos-def by adding to diagonal {}".format(padd)
+    )
+    return inv_he
+
+
+def force_pos_def(h):
+    """
+    from pricession lost hessian matrix
+    eigen value is small
+
+    dot(H, v[:,i] = e[i] v[:,i]
+    dot(H, v)[:,i] = e[i] v[:,i]
+    dot(inv(v), dot(H, v)) = diag(e)
+    H = dot(v, dot(diag(e), inv(v))
+
+    """
+    e, v = np.linalg.eig(h)
+    inv_he = np.linalg.pinv(h)
+    e_min = np.min(e)
+    e_max = np.max(np.abs(e))
+    if e_min > 0:
+        return inv_he
+    if e_min > -0.00001 * e_max:  # pricession cause small nagtive eigen value
+        idx = np.argmin(e)
+        e_new = np.where(e < 0, e - e_min * 1.1, e)
+        warnings.warn(
+            "Matrix forced pos-def by adding {} to eigen value".format(
+                -e_min * 1.1
+            )
+        )
+        h = np.dot(v, np.dot(np.diag(e_new), np.linalg.inv(v)))
+        inv_he = np.linalg.inv(h)
+        return inv_he
+    return force_pos_def_minuit2(inv_he)
+
+
+def cal_hesse_error(
+    fcn, params={}, check_posi_def=True, force_pos=True, save_npy=True
+):
     """
     This function calculates the errors of all trainable variables.
     The errors are given by the square root of the diagonal of the inverse Hessian matrix.
@@ -205,10 +282,15 @@ def cal_hesse_error(fcn, params={}, check_posi_def=True, save_npy=True):
         params
     )  # data_w,mcdata,weight=weights,batch=50000)
     print("Time for calculating errors:", time.time() - t)
-    if check_positive_definite(h.numpy()):
-        inv_he = np.linalg.inv(h.numpy())
+    h = h.numpy()
+    inv_he = np.linalg.inv(h)
+    if check_posi_def and check_positive_definite(inv_he):
+        inv_he = np.linalg.inv(h)
     else:
-        inv_he = np.linalg.pinv(h.numpy())
+        inv_he = np.linalg.pinv(h)
+        if force_pos:
+            inv_he = force_pos_def(h)
+
     if save_npy:
         np.save("error_matrix.npy", inv_he)
     diag_he = inv_he.diagonal()
@@ -216,6 +298,88 @@ def cal_hesse_error(fcn, params={}, check_posi_def=True, save_npy=True):
     return hesse_error, inv_he
 
 
+def cal_hesse_correct(fcn, params={}, corr_params={}, force_pos=True):
+    t = time.time()
+    nll, g, h = fcn.nll_grad_hessian(params)
+    # data_w,mcdata,weight=weights,batch=50000)
+    h = h.numpy()
+    var_names = fcn.vm.trainable_vars
+    params0 = fcn.get_params()
+    x0 = np.array([params0[k] for k in var_names])
+    idxs = [var_names.index(k) for k in corr_params]
+    # calculate hessian matrix for special params
+    _epsilon = 1e-3
+    _epsilon2 = _epsilon * 2
+    x = x0.copy()
+    for i in idxs:
+        for j in range(len(var_names)):
+            if j in idxs and i > j:
+                continue
+            if i == j:
+                x[i] += 2 * _epsilon
+                nll_pp = fcn(x)
+                x[i] -= _epsilon
+                nll_pm = fcn(x)
+                x[i] -= 2 * _epsilon
+                nll_mp = fcn(x)
+                x[i] -= _epsilon
+                nll_mm = fcn(x)
+                x[i] += 2 * _epsilon
+                gp = (nll_pp - nll_mp) / 3 / _epsilon
+                gm = (nll_mp - nll_mm) / 3 / _epsilon
+                new_hi = (gp - gm) / _epsilon
+                h[i, i] = new_hi
+            else:
+                x[i] += _epsilon
+                x[j] += _epsilon
+                nll_pp = fcn(x)
+                x[j] -= 2 * _epsilon
+                nll_pm = fcn(x)
+                x[j] += _epsilon
+                x[i] -= 2 * _epsilon
+                x[j] += _epsilon
+                nll_mp = fcn(x)
+                x[j] -= 2 * _epsilon
+                nll_mm = fcn(x)
+                x[j] += _epsilon
+                x[i] += _epsilon
+                gp = (nll_pp - nll_pm) / _epsilon2
+                gm = (nll_mp - nll_mm) / _epsilon2
+                new_hi = (gp - gm) / _epsilon2
+                h[i, j] = new_hi
+                h[j, i] = new_hi
+    print("Time for calculating errors:", time.time() - t)
+    return h
+
+
+def num_hess_inv_3point(fcn, params={}, _epsilon=5e-4):
+    """
+    This function calculates the errors of all trainable variables.
+    The errors are given by the square root of the diagonal of the inverse Hessian matrix.
+
+    :param model: Model.
+    :return hesse_error: List of errors.
+    :return inv_he: The inverse Hessian matrix.
+    """
+    var_names = fcn.vm.trainable_vars
+    old_params = fcn.get_params()
+    fcn(params)
+    f_g = fcn.vm.trans_fcn_grad(fcn.nll_grad)
+    hess = []
+    x0 = np.array(fcn.vm.get_all_val(True))
+    for i in range(len(var_names)):
+        x0[i] += _epsilon
+        _, g1 = f_g(x0)
+        x0[i] -= 2 * _epsilon
+        _, g2 = f_g(x0)
+        x0[i] += _epsilon
+        hess.append([(i - j) / 2 / _epsilon for i, j in zip(g1, g2)])
+    hess_inv = np.linalg.inv(hess)
+    fcn(old_params)
+    hess_inv = fcn.vm.trans_error_matrix(hess_inv, x0)
+    return hess_inv
+
+
 def gen_data(
     amp,
     Ndata,
