@@ -49,6 +49,7 @@ The full data structure provided is ::
 Inner nodes are named as tuple of particles.
 
 """
+import itertools
 
 import numpy as np
 
@@ -82,7 +83,7 @@ class CalAngleData(dict):
     def get_weight(self):
         if "weight" in self:
             return self["weight"]
-        return tf.ones(data_shape(self))
+        return tf.ones(data_shape(self), dtype=get_config("dtype"))
 
     def get_angle(self, decay, p):
         """ get hilicity angle of decay which product particle p"""
@@ -103,12 +104,21 @@ class CalAngleData(dict):
         data = data_to_numpy(self.get_mass(name))
         return Hist1D.histogram(data, bins=bins, **kwargs)
 
-    def savetxt(self, file_name, order=None):
+    def savetxt(
+        self, file_name, order=None, cp_trans=False, save_charge=False
+    ):
         if order is None:
             order = self.get_decay().outs
         pi = [data_to_numpy(self.get_momentum(i)) for i in order]
+        if cp_trans:
+            pi = [[i[0], -i[1], -i[2], -i[3]] for i in pi]
         pi = np.stack(pi).transpose((1, 0, 2)).reshape((-1, 4))
         np.savetxt(file_name, pi)
+        if save_charge:
+            np.savetxt(
+                file_name[::-1].replace(".", ".c", 1)[::-1],
+                self["charge_conjugation"],
+            )
 
 
 def struct_momentum(p, center_mass=True) -> dict:
@@ -287,7 +297,7 @@ def cal_helicity_angle(
                         ang["alpha"]
                     )
                     if i.core in r_matrix:
-                        r_matrix[j] = r * r_matrix[i.core] * b_matrix[i.core]
+                        r_matrix[j] = r * b_matrix[i.core] * r_matrix[i.core]
                     else:
                         r_matrix[j] = r
                 if len(i.outs) == 3:
@@ -344,11 +354,11 @@ def cal_angle_from_particle(
     # get base z axis
     p4 = data[decay_group.top]["p"]
     p3 = LorentzVector.vect(p4)
-    base_z = np.array([[0.0, 0.0, 1.0]]) + np.zeros_like(p3)
+    base_z = np.array([[0.0, 0.0, 1.0]]) + tf.zeros_like(p3)
     if random_z:
         p3_norm = Vector3.norm(p3)
-        mask = np.expand_dims(p3_norm < 1e-5, -1)
-        base_z = np.where(mask, base_z, p3)
+        mask = tf.expand_dims(p3_norm < 1e-5, -1)
+        base_z = tf.where(mask, base_z, p3)
     # calculate chain angle
     for i in decay_chain_struct:
         data_i = cal_helicity_angle(data, i, base_z=base_z)
@@ -483,8 +493,13 @@ def add_relative_momentum(data: dict):
     return data
 
 
+def parity_trans(p, charges):
+    charges = charges[: p.shape[0], None]
+    return tf.where(charges > 0, p, LorentzVector.neg(p))
+
+
 def prepare_data_from_decay(
-    fnames, decs, particles=None, dtype=None, **kwargs
+    fnames, decs, particles=None, dtype=None, charges=None, **kwargs
 ):
     """
     Transform 4-momentum data in files for the amplitude model automatically via DecayGroup.
@@ -500,6 +515,8 @@ def prepare_data_from_decay(
     if particles is None:
         particles = sorted(decs.outs)
     p = load_dat_file(fnames, particles, dtype=dtype)
+    if charges is not None:
+        p = {k: parity_trans(v, charges) for k, v in p.items()}
     data = cal_angle_from_momentum(p, decs, **kwargs)
     return data
 
@@ -543,6 +560,58 @@ def get_chain_data(data, decay_chain=None):
     return ret
 
 
+def cal_angle_from_momentum_base(
+    p,
+    decs: DecayGroup,
+    using_topology=True,
+    center_mass=False,
+    r_boost=True,
+    random_z=False,
+    batch=65000,
+) -> CalAngleData:
+    """
+    Transform 4-momentum data in files for the amplitude model automatically via DecayGroup.
+
+    :param p: 4-momentum data
+    :param decs: DecayGroup
+    :return: Dictionary of data
+    """
+    if data_shape(p) is None:
+        return cal_angle_from_momentum_single(
+            p, decs, using_topology, center_mass, r_boost, random_z
+        )
+    ret = []
+    for i in split_generator(p, batch):
+        ret.append(
+            cal_angle_from_momentum_single(
+                i, decs, using_topology, center_mass, r_boost, random_z
+            )
+        )
+    return data_merge(*ret)
+
+
+def identical_particles_swap(id_particles):
+    ret = []
+    for i in id_particles:
+        ret.append(list(itertools.permutations(i)))
+    for i in itertools.product(*ret):
+        yield i
+
+
+def identical_particles_swap_p(p4, id_particles):
+    for comb in identical_particles_swap(id_particles):
+        all_keys = tuple(p4.keys())
+        name_map = {str(k): k for k in all_keys}
+        swap_map = {}
+        for c, p_list in zip(comb, id_particles):
+            for ci, pi in zip(c, p_list):
+                swap_map[ci] = name_map[pi]
+        new_order = tuple([swap_map.get(str(i), i) for i in all_keys])
+        if all_keys == new_order:
+            continue
+        yield (new_order, comb), dict(zip(new_order, p4.values()))
+
+
 def cal_angle_from_momentum(
     p,
     decs: DecayGroup,
@@ -560,13 +629,19 @@ def cal_angle_from_momentum(
     :return: Dictionary of data
     """
     ret = []
-    for i in split_generator(p, batch):
-        ret.append(
-            cal_angle_from_momentum_single(
-                i, decs, using_topology, center_mass, r_boost, random_z
+    id_particles = decs.identical_particles
+    data = cal_angle_from_momentum_base(
+        p, decs, using_topology, center_mass, r_boost, random_z, batch
+    )
+    if id_particles is None or len(id_particles) == 0:
+        return data
+    else:
+        data["id_swap"] = {}
+        for i, pi in identical_particles_swap_p(p, id_particles):
+            data["id_swap"][i] = cal_angle_from_momentum_base(
+                pi, decs, using_topology, center_mass, r_boost, random_z, batch
             )
-        )
-    return data_merge(*ret)
+        return data
 
 
 def cal_angle_from_momentum_single(
