@@ -1,3 +1,4 @@
+import contextlib
 import copy
 import functools
 import itertools
@@ -24,10 +25,13 @@ from tf_pwa.amp import (
     get_particle,
 )
 from tf_pwa.applications import (
+    cal_hesse_correct,
     cal_hesse_error,
     corr_coef_matrix,
     fit,
     fit_fractions,
+    force_pos_def,
+    num_hess_inv_3point,
 )
 from tf_pwa.cal_angle import prepare_data_from_decay
 from tf_pwa.data import (
@@ -41,11 +45,12 @@ from tf_pwa.data import (
 )
 from tf_pwa.fit import FitResult
 from tf_pwa.fit_improve import minimize as my_minimize
-from tf_pwa.model import FCN, CombineFCN, Model, Model_new
-from tf_pwa.model.cfit import Model_cfit
+from tf_pwa.model import FCN, CombineFCN, MixLogLikehoodFCN, Model, Model_new
+from tf_pwa.model.cfit import Model_cfit, Model_cfit_cached
 from tf_pwa.model.opt_int import ModelCachedAmp, ModelCachedInt
 from tf_pwa.particle import split_particle_type
 from tf_pwa.root_io import has_uproot, save_dict_to_root
+from tf_pwa.tensorflow_wrapper import tf
 from tf_pwa.utils import time_print
 from tf_pwa.variable import Variable, VarsManager
 
@@ -221,6 +226,11 @@ class ConfigLoader(BaseConfig):
         self.amps[vm] = amp
         return amp
 
+    def eval_amplitude(self, *p):
+        data = self.data.cal_angle(p)
+        amp = self.get_amplitude()
+        return amp(data)
+
     def check_valid_jp(self, decay_group):
         for decay_chain in decay_group:
             for dec in decay_chain:
@@ -307,9 +317,9 @@ class ConfigLoader(BaseConfig):
                     "m_": "mass_",
                     "g_": "width_",
                 }
-                particle_config = self.config["particle"][i].copy()
+                particle_config = self.decay_config.particle_property[i]
 
-                params_dic = self.config["particle"][i].get("params", None)
+                params_dic = particle_config.get("params", None)
                 if params_dic is None:
                     params_dic = {}
                 for name in list(particle_config):
@@ -349,20 +359,17 @@ class ConfigLoader(BaseConfig):
                 if isinstance(p_i.mass, Variable) or isinstance(
                     p_i.width, Variable
                 ):
-                    if (
-                        "float" in self.config["particle"][i]
-                        and self.config["particle"][i]["float"]
-                    ):
-                        if "m" in self.config["particle"][i]["float"]:
+                    if "float" in particle_config and particle_config["float"]:
+                        if "m" in particle_config["float"]:
                             p_i.mass.freed()  # set_fix(i+'_mass',unfix=True)
-                            if "m_max" in self.config["particle"][i]:
-                                upper = self.config["particle"][i]["m_max"]
+                            if "m_max" in particle_config:
+                                upper = particle_config["m_max"]
                             # elif m_sigma is not None:
                             #    upper = self.config["particle"][i]["m0"] + 10 * m_sigma
                             else:
                                 upper = None
-                            if "m_min" in self.config["particle"][i]:
-                                lower = self.config["particle"][i]["m_min"]
+                            if "m_min" in particle_config:
+                                lower = particle_config["m_min"]
                             # elif m_sigma is not None:
                             #    lower = self.config["particle"][i]["m0"] - 10 * m_sigma
                             else:
@@ -370,16 +377,16 @@ class ConfigLoader(BaseConfig):
                             self.bound_dic[str(p_i.mass)] = (lower, upper)
                         else:
                             self._neglect_when_set_params.append(str(p_i.mass))
-                        if "g" in self.config["particle"][i]["float"]:
+                        if "g" in particle_config["float"]:
                             p_i.width.freed()  # amp.vm.set_fix(i+'_width',unfix=True)
-                            if "g_max" in self.config["particle"][i]:
-                                upper = self.config["particle"][i]["g_max"]
+                            if "g_max" in particle_config:
+                                upper = particle_config["g_max"]
                             # elif g_sigma is not None:
                             #    upper = self.config["particle"][i]["g0"] + 10 * g_sigma
                             else:
                                 upper = None
-                            if "g_min" in self.config["particle"][i]:
-                                lower = self.config["particle"][i]["g_min"]
+                            if "g_min" in particle_config:
+                                lower = particle_config["g_min"]
                             # elif g_sigma is not None:
                             #    lower = self.config["particle"][i]["g0"] - 10 * g_sigma
                             else:
@@ -398,8 +405,8 @@ class ConfigLoader(BaseConfig):
                         )  # p_i.width.name
 
                 # share helicity variables
-                if "coef_head" in self.config["particle"][i]:
-                    coef_head = self.config["particle"][i]["coef_head"]
+                if "coef_head" in particle_config:
+                    coef_head = particle_config["coef_head"]
                     if coef_head in res_dec:
                         d_coef_head = res_dec[coef_head]
                         for j, h in zip(d, d_coef_head):
@@ -410,7 +417,7 @@ class ConfigLoader(BaseConfig):
                         # share total radium
                         d_coef_head.total.r_shareto(d.total)
                     else:
-                        self.config["particle"][coef_head]["coef_head"] = i
+                        particle_config["coef_head"] = i
 
         equal_params = dic.get("equal", {})
         for k, v in equal_params.items():
@@ -439,7 +446,14 @@ class ConfigLoader(BaseConfig):
             if not isinstance(w_bkg, list):
                 w_bkg = [w_bkg]
             for wb in w_bkg:
-                model.append(Model_cfit(amp, wb, bg_function, eff_function))
+                if self.config["data"].get("cached_amp", False):
+                    model.append(
+                        Model_cfit_cached(amp, wb, bg_function, eff_function)
+                    )
+                else:
+                    model.append(
+                        Model_cfit(amp, wb, bg_function, eff_function)
+                    )
         elif "inmc" in self.config["data"]:
             float_wmc = self.config["data"].get(
                 "float_inmc_ratio_in_pdf", False
@@ -506,6 +520,20 @@ class ConfigLoader(BaseConfig):
             bg = [None] * self._Ngroup
         model = self._get_model(vm=vm, name=name)
         fcns = []
+        # print(self.config["data"].get("using_mix_likelihood", False))
+        if self.config["data"].get("using_mix_likelihood", False):
+            print("  Using Mix Likelihood")
+            fcn = MixLogLikehoodFCN(
+                model,
+                data,
+                phsp,
+                bg=bg,
+                batch=batch,
+                gauss_constr=self.gauss_constr_dic,
+            )
+            if all_data is None:
+                self.cached_fcn[vm] = fcn
+            return fcn
         for md, dt, mc, sb, ij in zip(model, data, phsp, bg, inmc):
             if self.config["data"].get("model", "auto") == "cfit":
                 fcns.append(
@@ -579,6 +607,7 @@ class ConfigLoader(BaseConfig):
             fcn = self.get_fcn(batch=batch)
         else:
             fcn = self.get_fcn([data, phsp, bg, inmc], batch=batch)
+        # print("sss")
         amp = self.get_amplitude()
         print("decay chains included: ")
         for i in self.full_decay:
@@ -626,9 +655,19 @@ class ConfigLoader(BaseConfig):
         inmc=None,
         batch=10000,
         using_cached=False,
+        method=None,
+        force_pos=True,
+        correct_params=None,
     ):
+        """
+        calculate parameters error
+        """
         if params is None:
             params = {}
+        if correct_params is None:
+            correct_params = []
+            if method is None:
+                method = "correct"
         if data is None:
             data, phsp, bg, inmc = self.get_all_data()
         if hasattr(params, "params"):
@@ -636,9 +675,25 @@ class ConfigLoader(BaseConfig):
         fcn = self.get_fcn([data, phsp, bg, inmc], batch=batch)
         if using_cached and self.inv_he is not None:
             hesse_error = np.sqrt(np.fabs(self.inv_he.diagonal())).tolist()
+        elif method == "3-point":
+            self.inv_he = num_hess_inv_3point(fcn, params)
+            diag_he = self.inv_he.diagonal()
+            hesse_error = np.sqrt(np.fabs(diag_he)).tolist()
+        elif method == "correct":
+            h = cal_hesse_correct(fcn, params, correct_params)
+            if force_pos:
+                self.inv_he = force_pos_def(h)
+            else:
+                self.inv_he = np.linalg.pinv(h)
+            diag_he = self.inv_he.diagonal()
+            hesse_error = np.sqrt(np.fabs(diag_he)).tolist()
         else:
             hesse_error, self.inv_he = cal_hesse_error(
-                fcn, params, check_posi_def=True, save_npy=True
+                fcn,
+                params,
+                check_posi_def=True,
+                save_npy=True,
+                force_pos=force_pos,
             )
         # print("parameters order")
         # print(fcn.model.Amp.vm.trainable_vars)
@@ -807,11 +862,37 @@ class ConfigLoader(BaseConfig):
         return True
 
     def save_params(self, file_name):
-        params = self.get_params(False)
+        params = self.get_params()
         val = {k: float(v) for k, v in params.items()}
         with open(file_name, "w") as f:
             json.dump(val, f, indent=2)
 
+    @contextlib.contextmanager
+    def params_trans(self):
+        with self.vm.error_trans(self.inv_he) as f:
+            yield f
+
+    def save_tensorflow_model(self, dir_name):
+        class CustomModule(tf.Module):
+            def __init__(self, config_name, share_dict, final_params):
+                self.config = ConfigLoader(config_name, share_dict=share_dict)
+                self.amp = self.config.get_amplitude()
+                self.config.set_params(final_params)
+                self.all_variables = self.amp.vm.variables
+
+            @tf.function()
+            def __call__(self, *p):
+                data = self.config.data.cal_angle(p)
+                return self.amp(data)
+
+        module = CustomModule(self.config, self.share_dict, self.get_params())
+        n_p = len(self.get_dat_order())
+        input_p = [tf.TensorSpec([None, 4], tf.float64) for i in range(n_p)]
+        call = module.__call__.get_concrete_function(*input_p)
+        tf.saved_model.save(
+            module, dir_name, signatures={"serving_default": call}
+        )
+
 
 def set_prefix_constrains(vm, base, params_dic, self):
     prefix = base.get_variable_name()
