@@ -113,6 +113,7 @@ def sum_hessian(
                 )
             g_i = tape.gradient(y_i, var, unconnected_gradients="zero")
         h_s_i = []
+        # h_s_i  = tape0.jacobian(tf.stack(g_i), var, unconnected_gradients="zero")
         for gi in g_i:
             # 2nd order derivative
             h_s_i.append(tape0.gradient(gi, var, unconnected_gradients="zero"))
@@ -128,6 +129,52 @@ def sum_hessian(
     return nll, g, h
 
 
+def sum_grad_hessp(
+    f,
+    p,
+    data,
+    var,
+    weight=1.0,
+    trans=tf.identity,
+    resolution_size=1,
+    args=(),
+    kwargs=None,
+):
+    """
+    The parameters are the same with ``sum_gradient()``, but this function will return hessian as well,
+    which is the matrix of the second-order derivative.
+
+    :return: Real number NLL, list gradient, 2-D list hessian
+    """
+    kwargs = kwargs if kwargs is not None else {}
+    if isinstance(weight, float):
+        weight = _loop_generator(weight)
+    y_s = []
+    g_s = []
+    h_s = []
+    from tensorflow.python.eager import forwardprop
+
+    for data_i, weight_i in zip(data, weight):
+        with forwardprop.ForwardAccumulator(var, list(p)) as acc:
+            with tf.GradientTape() as tape:
+                y_i = _batch_sum(
+                    f, data_i, weight_i, trans, resolution_size, args, kwargs
+                )
+            g_i = tape.gradient(y_i, var, unconnected_gradients="zero")
+        hessp = acc.jvp(g_i, unconnected_gradients="zero")
+        y_s.append(y_i)
+        g_s.append(g_i)
+        h_s.append(hessp)
+        # print(hessp)
+    nll = tf.reduce_sum(y_s)
+    # print("ll: ", nll)
+    g = tf.reduce_sum(g_s, axis=0)
+    h = tf.reduce_sum(h_s, axis=0)
+    # print(h)
+    # h = [[sum(j) for j in zip(*i)] for i in h_s]
+    return nll, g, h
+
+
 def sum_gradient_new(
     amp,
     data,
@@ -267,6 +314,30 @@ class BaseModel(object):
             tf.reduce_sum(weight * ln_data) - sw * tf.math.log(int_mc)
         )
 
+    def sum_nll_grad_bacth(self, data):
+        weight = [i.get("weight", tf.ones((data_shape(i),))) for i in data]
+        ln_data, g_ln_data = sum_gradient(
+            self.signal,
+            data,
+            self.signal.trainable_variables,
+            weight=weight,
+            trans=clip_log,
+            resolution_size=self.resolution_size,
+        )
+        return -ln_data, [-i for i in g_ln_data]
+
+    def sum_log_integral_grad_batch(self, mcdata, ndata):
+        mc_weight = [i["weight"] for i in mcdata]
+        int_mc, g_int_mc = sum_gradient(
+            self.signal,
+            mcdata,
+            self.signal.trainable_variables,
+            weight=mc_weight,
+        )
+        return tf.math.log(int_mc) * ndata, [
+            ndata / int_mc * i for i in g_int_mc
+        ]
+
     def nll_grad(self, data, mcdata, batch=65000):
         weight = data.get("weight", tf.ones((data_shape(data),)))
         weight_rw = tf.reduce_sum(
@@ -347,6 +418,65 @@ class BaseModel(object):
         nll = -ln_data + sw * tf.math.log(int_mc)
         return nll, g
 
+    def grad_hessp_batch(self, p, data, mcdata, weight, mc_weight):
+        """
+        ``self.nll_grad()`` is replaced by this one???
+
+        .. math::
+          - \\frac{\\partial \\ln L}{\\partial \\theta_k } =
+            -\\sum_{x_i \\in data } w_i \\frac{\\partial}{\\partial \\theta_k} \\ln f(x_i;\\theta_k)
+            + (\\sum w_j ) \\left( \\frac{ \\partial }{\\partial \\theta_k} \\sum_{x_i \\in mc} f(x_i;\\theta_k) \\right)
+              \\frac{1}{ \\sum_{x_i \\in mc} f(x_i;\\theta_k) }
+
+        :param data:
+        :param mcdata:
+        :param weight:
+        :param mc_weight:
+        :return:
+        """
+        if not hasattr(self, "hess_product_vector_i"):
+            self.hess_product_vector_i = [tf.Variable(i) for i in p]
+        for i, j in zip(self.hess_product_vector_i, p):
+            i.assign(j)
+        weight = list(weight)
+        sw = tf.reduce_sum([tf.reduce_sum(i) for i in weight])
+        ln_data, g_ln_data, hessp_ln_data = sum_grad_hessp(
+            self.signal,
+            self.hess_product_vector_i,
+            data,
+            self.signal.trainable_variables,
+            weight=weight,
+            trans=clip_log,
+            resolution_size=self.resolution_size,
+        )
+
+        # print("hessp_ln_data",hessp_ln_data)
+
+        int_mc, g_int_mc, hessp_int_mc = sum_grad_hessp(
+            self.signal,
+            self.hess_product_vector_i,
+            mcdata,
+            self.signal.trainable_variables,
+            weight=mc_weight,
+        )
+
+        # print("hessp_int_mc", hessp_int_mc)
+
+        sw = tf.cast(sw, ln_data.dtype)
+
+        g = list(
+            map(lambda x: -x[0] + sw * x[1] / int_mc, zip(g_ln_data, g_int_mc))
+        )
+
+        g_int_mc = np.array(g_int_mc)
+        hessp2 = sw * (
+            hessp_int_mc / int_mc
+            - g_int_mc * np.dot(p, g_int_mc) / int_mc ** 2
+        )
+        # print("hessp2", hessp2)
+        # print("ret", g, hessp2 - hessp_ln_data)
+        return g, hessp2 - hessp_ln_data
+
     def nll_grad_hessian(self, data, mcdata, batch=25000):
         """
         The parameters are the same with ``self.nll()``, but it will return Hessian as well.
@@ -462,6 +592,19 @@ class Model(object):
             return data, alpha * weight
         return data, weight
 
+    def mix_data_bakcground(self, data, bg):
+        ret, weight = self.get_weight_data(
+            data, weight=data.get_weight(), bg=bg, alpha=True
+        )
+        ret["weight"] = weight
+        return ret
+
+    def sum_nll_grad_bacth(self, data):
+        return self.model.sum_nll_grad_bacth(data)
+
+    def sum_log_integral_grad_batch(self, mcdata, ndata):
+        return self.model.sum_log_integral_grad_batch(mcdata, ndata)
+
     def nll(
         self,
         data,
@@ -521,6 +664,29 @@ class Model(object):
             batch=batch,
         )
 
+    # @tf.function
+    def grad_hessp_batch(self, p, data, mcdata, weight, mc_weight):
+        """
+        ``self.nll_grad()`` is replaced by this one???
+
+        .. math::
+          - \\frac{\\partial \\ln L}{\\partial \\theta_k } =
+            -\\sum_{x_i \\in data } w_i \\frac{\\partial}{\\partial \\theta_k} \\ln f(x_i;\\theta_k)
+            + (\\sum w_j ) \\left( \\frac{ \\partial }{\\partial \\theta_k} \\sum_{x_i \\in mc} f(x_i;\\theta_k) \\right)
+              \\frac{1}{ \\sum_{x_i \\in mc} f(x_i;\\theta_k) }
+
+        :param data:
+        :param mcdata:
+        :param weight:
+        :param mc_weight:
+        :return:
+        """
+        data_i = ({**i, "weight": j} for i, j in zip(data, weight))
+        mcdata_i = ({**i, "weight": j} for i, j in zip(mcdata, mc_weight))
+        return self.model.grad_hessp_batch(
+            p, data_i, mcdata_i, weight, mc_weight
+        )
+
     # @tf.function
     def nll_grad_batch(self, data, mcdata, weight, mc_weight):
         """
@@ -1033,6 +1199,26 @@ class FCN(object):
         constr_hessian = self.gauss_constr.get_constrain_hessian()
         return nll + constr, g + constr_grad, h + constr_hessian
 
+    @time_print
+    def grad_hessp(self, x, p, batch=None):
+        if batch is None:
+            batch = self.batch
+        g, h = self.get_grad_hessp(x, p, batch)
+        constr_grad = self.gauss_constr.get_constrain_grad()
+        constr_hessian = 0.0  # self.gauss_constr.get_constrain_hessp(p)
+        return g + constr_grad, h + constr_hessian
+
+    def get_grad_hessp(self, x, p, batch):
+        self.model.set_params(x)
+        grad, hessp = self.model.grad_hessp_batch(
+            p,
+            self.batch_data,
+            self.batch_mcdata,
+            weight=list(data_split(self.weight, self.batch)),
+            mc_weight=self.batch_mc_weight,
+        )
+        return grad, hessp
+
 
 class CombineFCN(object):
     """
@@ -1156,3 +1342,90 @@ class CombineFCN(object):
         constr_grad = self.gauss_constr.get_constrain_grad()
         constr_hessian = self.gauss_constr.get_constrain_hessian()
         return nll + constr, g + constr_grad, h + constr_hessian
+
+    def get_grad_hessp(self, x, p, batch):
+        """
+        :param x: List. Values of variables.
+        :return nll: Real number. The value of NLL.
+        :return gradients: List of real numbers. The gradients for each variable.
+        """
+        hs = []
+        gs = []
+        for i in self.fcns:
+            g, h = i.get_grad_hessp(x, p, batch)
+            hs.append(h)
+            gs.append(g)
+        return tf.reduce_sum(gs, axis=0), tf.reduce_sum(hs, axis=0)
+
+    def grad_hessp(self, x, p, batch=None):
+        grad, hessp = self.get_grad_hessp(x, p, batch)
+        constr_grad = self.gauss_constr.get_constrain_grad()
+        return grad + constr_grad, hessp
+
+
+class MixLogLikehoodFCN(CombineFCN):
+
+    """
+    This class implements methods to calculate the NLL as well as its derivatives for a general function.
+
+    :param model: List of model object.
+    :param data: List of data array.
+    :param mcdata: list of MCdata array.
+    :param bg: list of Background array.
+    :param batch: The length of array to calculate as a vector at a time. How to fold the data array may depend on the GPU computability.
+    """
+
+    def __init__(
+        self,
+        model,
+        data,
+        mcdata,
+        bg=None,
+        batch=65000,
+        gauss_constr={},
+    ):
+        self.cached_nll = 0.0
+        assert model is not None, "model required"
+        assert data is not None, "data required"
+        assert mcdata is not None, "mcdata required"
+        self.fcns = []
+        self.cached_nll = 0.0
+        if bg is None:
+            bg = _loop_generator(None)
+        self.datas = []
+        self.weight_phsps = []
+        self.n_datas = []
+        self.model = model
+        for model_i, data_i, mcdata_i, bg_i in zip(model, data, mcdata, bg):
+            data_s = model_i.mix_data_bakcground(data_i, bg_i)
+            self.datas.append(data_s)
+            weight_phsp = type(mcdata_i)(
+                {k: v for k, v in mcdata_i.items()}
+            )  # simple copy
+            w = weight_phsp.get_weight()
+            weight_phsp["weight"] = w / tf.reduce_sum(w)
+            self.n_datas.append(tf.reduce_sum(data_s.get_weight()))
+            self.weight_phsps.append(list(split_generator(weight_phsp, batch)))
+            self.fcns.append(FCN(model_i, data_i, mcdata_i, bg_i))
+        self.data_merge = list(split_generator(data_merge(*self.datas), batch))
+        self.vm = self.model[0].vm
+        self.gauss_constr = GaussianConstr(self.vm, gauss_constr)
+
+    def get_nll_grad(self, x={}):
+        """
+        :param x: List. Values of variables.
+        :return nll: Real number. The value of NLL.
+        :return gradients: List of real numbers. The gradients for each variable.
+        """
+        [i.set_params(x) for i in self.model]
+        nlls = []
+        gs = []
+        nll, g = self.model[0].sum_nll_grad_bacth(self.data_merge)
+        nlls.append(nll)
+        gs.append(g)
+        for i, k, l in zip(self.model, self.weight_phsps, self.n_datas):
+            nll, g = i.sum_log_integral_grad_batch(k, l)
+            nlls.append(nll)
+            gs.append(g)
+        print(sum(nlls))
+        return sum(nlls), tf.reduce_sum(gs, axis=0)
